/* UNH CS 758 assignment 3

   Record trigram probabilities from input and generate text.

   see babble.h for overview of major data structures
*/

#include <stdio.h>
#include <stdlib.h>  /* free, EXIT_SUCCESS, rand */
#include <ctype.h>   /* isalnum */
#include <string.h>  /* strcpy */
#include <time.h>    /* time */
#include "utils.h"
#include "babble.h"


/* declare functions used in main but defined elsewhere */
dict *make_array_dict(void);
dict *make_ll_dict(void);
dict *make_ht_bad_dict(void);
dict *make_ht_add3_dict(void);
dict *make_ht_kr_dict(void);
dict *make_ht_tab_dict(void);


/* 30 was chosen after looking at
   https://math.wvu.edu/~hdiamond/Math222F17/Sigurd_et_al-2004-Studia_Linguistica.pdf */
#define MAX_TOK_LEN 30
#define START_TOK "$START$"
#define TOK_SEPARATOR '\t'

#define DEFAULT_SEED 31415


/************************************************************
 * read input
 ************************************************************/


int is_standalone(char c)
{
  if (strchr(".?!,;:'\"", c) == NULL)
    return 0;
  else
    return 1;
}


int is_tokchar(char c)
{
  return (isalnum(c) || is_standalone(c));
}


/* anything not alphanum or "standalone" (as defined above) is treated
   as whitespace.  standalone chars each form their own token.
   ignores leading whitespace.  returns 0 if no more input to read,
   otherwise token length */
int read_token(FILE *infile, char *buf, int bufsize)
{
  int i = 0;
  char c;

  if (bufsize <= 2) panic("buffer too small in read_token");
  do {
    c = fgetc(infile);
  /* skip leading whitespace */
  } while (! ((c == EOF) || (is_tokchar(c))));
  if (is_standalone(c)) {
    buf[i++] = c;
  } else if (isalnum(c)) {
    do {
      buf[i++] = tolower(c);
    } while ((i < (bufsize-1)) && (c = fgetc(infile), isalnum(c)));
    if (is_tokchar(c))
      if (ungetc(c, infile) == EOF) panic("error ungetting char in read_token");
  }
  if (ferror(infile)) panic("error reading tokens");
  buf[i]='\0';
  /* fprintf(stderr, "got token '%s' (%d)\n", buf, i); */
  return i;
}


/************************************************************
 * distributions
 ************************************************************/


dist *make_dist(void)
{
  count *cs;
  dist *dst;
  
  cs = (count *) malloc(sizeof(count) * DIST_INIT_SIZE);
  if (!cs) panic("failed to malloc counts for a new dist");
  dst = (dist *) malloc(sizeof(dist));
  if (!dst) panic("failed to malloc new dist");
  dst->size = DIST_INIT_SIZE;
  dst->len = 0;
  dst->counts = cs;
  return dst;
}


void free_dist(dist *dst)
{
  int i;
  count *cs = dst->counts;

  for (i = 0; i < dst->len; i++) {
    free(cs[i].token);
  }
  free(cs);
  free(dst);
}


void fprint_dist(FILE *outfile, dist *dst)
{
  int i;
  
  fprintf(outfile, "   dist of size %d with %d entries\n", dst->size, dst->len);
  for (i = 0; i < dst->len; i++) {
    count c = dst->counts[i];
    
    fprintf(outfile, "      tok `%s' \t%d counts\n", c.token, c.count);
  }
}


/* returns nonzero if failed to find tok, otherwise sets c */
int find_token(dist *dst, char *tok, count **c)
{
  int i;
  count *cs = dst->counts;

  for (i = 0; i < dst->len; i++) {
    if (strcmp(tok, cs[i].token) == 0) {
      *c = &(cs[i]);
      return 0;
    }
  }
  return 1;;
}


/* initializes count of tok to 1.  does not check that tok is not
   already in dst. does not copy tok. */
void insert_count(dist *dst, char *tok)
{
  count *c;

  if (dst->len == dst->size) {
    int newsize = dst->size * 2;
    dst->counts = (count *) realloc(dst->counts, sizeof(count) * newsize);
    if (! dst->counts) panic("failed to increase size of counts in dist");
    dst->size = newsize;
  }
  c = &((dst->counts)[dst->len++]);
  c->token = tok;
  c->count = 1;
  /* c->to_right will contain garbage for now, which is fine */
}


/* sets the to_right fields of all the counts, in preparation for
   generation */
void set_dist_to_rights(char *key, dist *dst)
{
  int accum = 0;
  int i;

  for (i = (dst->len - 1); i >= 0; i--) {
    count *c = &(dst->counts)[i];
    c->to_right = accum;
    accum += c->count;
  }
}


/************************************************************
 * build model
 ************************************************************/


/* note that the returned key is a static buffer that will be
   overwritten the next time this function is called */
char *form_key(char *tok1, char *tok2)
{
  static char key[(2 * MAX_TOK_LEN) + 2];
  int i, j;
  
  for (i = 0; tok1[i] != '\0'; i++) {
    key[i] = tok1[i];
  };
  key[i++] = TOK_SEPARATOR;
  for (j = 0; tok2[j] != '\0'; i++, j++) {
    key[i] = tok2[j];
  };
  key[i] = '\0';
  return key;
}


/* if key or tok not found, new ones for dict will be
   heap-allocated */
void incr_tok(dict *d, char *key, char *tok)
{
  dist *dst;
  count *c;

  /* fprintf(stderr, "incrementing `%s'->'%s'\n", key, tok); */
  if (d->lookup(d->data, key, &dst)) {
    /* key not found, allocate new key and dist */
    dst = make_dist();
    if (d->insert(d->data, strcpy_new(key), dst)) 
      panic("inserting into dictionary failed");
  }
  if (find_token(dst, tok, &c)) {
    /* val not found in dst, allocate new */
    insert_count(dst, strcpy_new(tok));
  } else {
    c->count++;
  }
}


void free_dict_pair(char *key, dist *val)
{
  free(key);
  free_dist(val);
}


void free_dict (dict *d)
{
  d->iter(d->data, &free_dict_pair);
  d->free(d->data);
  free(d);
}


/* for use by the fprint function defined for each dict implementation */
void fprint_pair(FILE *outfile, char *key, dist *val)
{
  fprintf(outfile, "key '%s' ->\n", key);
  fprint_dist(outfile, val);
}


void build_model(dict *d, FILE *infile)
{
  /* stack-allocated input buffers */
  char buf1[MAX_TOK_LEN];
  char buf2[MAX_TOK_LEN];
  char buf3[MAX_TOK_LEN];
  char *tok1 = buf1, *tok2 = buf2, *tok3 = buf3;

  strncpy(tok1, START_TOK, MAX_TOK_LEN);
  strncpy(tok2, START_TOK, MAX_TOK_LEN);
  while (read_token(infile, tok3, MAX_TOK_LEN) > 0) {
    char *key = form_key(tok1, tok2);
    char *tmp;
    
    incr_tok(d, key, tok3);
    tmp = tok1;
    tok1 = tok2;
    tok2 = tok3;
    tok3 = tmp;
  }
}


/************************************************************
 * generate text
 ************************************************************/


/* given an array of count structs, find the target in the cumulative
   distribution represented by the to_right fields.  uses binary
   search between left and right inclusive.  invariant: the target count
   is in a bucket within the left-right (inclusive) range. */
char *find_tok(count *cs, int target, int left, int right)
{
  if (left == right)
    return cs[left].token;
  else {
    int mid = (left + right) / 2;
    count c = cs[mid];
    int mid_min = c.to_right + 1;
    int mid_max = c.to_right + c.count;
    if (target > mid_max) {
      return find_tok(cs, target, left, mid-1);
    } else if (mid_min > target) {
      return find_tok(cs, target, mid+1, right);
    } else
      return c.token;
  }
}

   
char *sample_tok(dist *dst)
{
  int total_counts;
  int target;

  if (dst->len < 1) panic("dist with less than 1 count");
  total_counts = dst->counts[0].count + dst->counts[0].to_right;
  target = (rand() % total_counts) + 1;
  return find_tok(dst->counts, target, 0, dst->len-1);
}


void generate(dict *d, FILE *outfile, int nwords)
{
  char *tok1 = START_TOK;
  char *tok2 = START_TOK;
  int i;
  
  d->iter(d->data, set_dist_to_rights);
  for (i = 0; i < nwords; i++) {
    char *key = form_key(tok1, tok2);
    dist *dst;
    char *tok3;

    if (d->lookup(d->data, key, &dst)) {
      /* nothing has ever followed tok1 and tok2 in the input text.
	 this can happen only if they are the last two words of the
	 input.  to generate more text, we restart the model. */
      tok1 = START_TOK;
      tok2 = START_TOK;
      key = form_key(tok1, tok2);
      if (d->lookup(d->data, key, &dst))
	panic("failed to restart the model");
    }
    tok3 = sample_tok(dst);
    fprintf(outfile, " %s", tok3);
    tok1 = tok2;
    tok2 = tok3;
  }
  /* fprintf(outfile, "\n"); */
}


/************************************************************
 * arg parsing and main

 babble hashtable add3 50 c < ../data/twogentlemenlebowski.txt

 ************************************************************/


void seed_rand(int seed)
{
  if (seed < 1)
    srand(time(NULL));
  else
    srand(seed);
}


void usage(void)
{
    fprintf(stderr, "usage: babble [-seed <num>] <ds> [<hf> [c]] <n>\n");
    fprintf(stderr,
	    "where <ds> is one of: array, linkedlist, ht\n"
	    "   if <ds>=ht, <hf> is one of: bad, add3, kr, tab\n"
	    "   if <ds>=ht, optional arg c indicates dumping bucket counts\n"
	    "and <n> is the number of words of text to generated.\n"
	    "Text is read from stidin and generated to stdout.\n");
    exit(EXIT_FAILURE);
}


void parse_args(int argc, char *const argv[],
		dict **d, int *seed, int *dump, int *n)
{
  int i;
  
  if ((argc < 3) || (argc > 7)) usage();
  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-seed") == 0) {
      /* advance to seed value */
      if ((i+1) == argc) usage();
      *seed = atoi(argv[++i]);
    } else if (strcmp(argv[i], "array") == 0)
      *d = make_array_dict();
    else if (strcmp(argv[i], "linkedlist") == 0)
      *d = make_ll_dict();      
    else if (strcmp(argv[i], "hashtable") == 0) {
      /* advance to hash function */
      if (++i == argc) usage();
      if (strcmp(argv[i], "bad") == 0)
	*d = make_ht_bad_dict();
      else if (strcmp(argv[i], "add3") == 0)
	*d = make_ht_add3_dict();
      else if (strcmp(argv[i], "kr") == 0)
	*d = make_ht_kr_dict();
      else if (strcmp(argv[i], "tab") == 0)
	*d = make_ht_tab_dict();
      else
	usage();
    } else if (strcmp(argv[i], "c") == 0)
      /* NOTE THAT HARNESS PASSES c LAST, contrary to usage() */
      *dump = TRUE;
    else if (isdigit(argv[i][0]))
      *n = atoi(argv[i]);
    else
      usage();
  }
}


int main(int argc, char *const argv[])
{
  dict *d;
  int seed = DEFAULT_SEED;
  int dump = FALSE;
  int nwords = 50;

  parse_args(argc, argv, &d, &seed, &dump, &nwords);
  seed_rand(seed);
  build_model(d, stdin);
  /* harness wants counts to stdout and babble to stderr! */
  if (dump) d->fprint(stdout, d->data);
  generate(d, stderr, nwords);
  free_dict(d);
  return EXIT_SUCCESS;
}


/* EOF */
